==> http_tools_abm-5.4.65135.tar.gz: http_tools_abm-5.4.65135/http_tools/answer.py <==
import dataclasses
import logging
import urllib.parse
from typing import Any, Optional, Dict, Union

from aiohttp import Payload
from aiohttp.web_response import Response
from dynamic_types.class_name import _prepare_class_name
from dynamic_types.create import create_type, create_dataclass
from init_helpers import custom_dumps
from init_helpers.dict_to_dataclass import dict_to_dataclass, convert_to_type
from yarl import URL

from .charset import Charset
from .http_status_codes import HttpStatusCode
from .mime_types import ContentType

logger = logging.getLogger(__name__)


class Nothing:
    pass


class ErrorCode(int):
    pass


_FIXED_CONTENT_TYPE__NAME = "_fixed_content_type"
_FIXED_STATUS_CODE__NAME = "_fixed_status_code"
_FIXED_PAYLOAD_TYPE__NAME = "_fixed_payload_type"
_FIXED_ERROR_CODE__NAME = "_fixed_error_code_type"


class Answer(Response):
    def __init__(self,
                 payload: Union[str, bytes, None],
                 status: HttpStatusCode = None,
                 content_type: Optional[Union[ContentType, str]] = None,
                 headers: Optional[Dict[str, str]] = None,
                 charset: Charset = None,
                 ):
        if status is None and hasattr(self, _FIXED_STATUS_CODE__NAME):
            status = getattr(self, _FIXED_STATUS_CODE__NAME)

        status_code = status or self.get_class_content_type()
        content_type = content_type or self.get_class_content_type()
        content_type = content_type.value if isinstance(content_type, ContentType) else content_type
        super().__init__(body=payload, status=status_code, content_type=content_type, headers=headers, charset=charset)

    @property
    def bytes_body(self) -> bytes:
        return self.body._value if isinstance(self.body, Payload) else self.body

    def __class_getitem__(cls, modifier: Union[HttpStatusCode, ContentType]) -> type:
        field_name_to_value = {}
        if isinstance(modifier, HttpStatusCode):
            if cls.get_class_status_code():
                raise TypeError("Status code already fixed")
            field_name_to_value[_FIXED_STATUS_CODE__NAME] = modifier
        elif isinstance(modifier, ContentType):
            if cls.get_class_content_type():
                raise TypeError("Content type already fixed")
            field_name_to_value[_FIXED_CONTENT_TYPE__NAME] = modifier
        else:
            raise TypeError(f"Unexpected modifier type: {type(modifier)}")

        new_type = create_type(_prepare_class_name([cls], modifier), [cls], field_name_to_value)
        # new_type = type(_prepare_class_name([cls], modifier), (cls, ), field_name_to_value)
        return new_type

    @classmethod
    def get_class_content_type(cls) -> Optional[ContentType]:
        if hasattr(cls, _FIXED_CONTENT_TYPE__NAME):
            return getattr(cls, _FIXED_CONTENT_TYPE__NAME)

    @classmethod
    def get_class_status_code(cls) -> Optional[HttpStatusCode]:
        if hasattr(cls, _FIXED_STATUS_CODE__NAME):
            return getattr(cls, _FIXED_STATUS_CODE__NAME)

    @classmethod
    def get_class_payload_type(cls) -> type:
        return bytes


class JsonableAnswer(Answer):
    """
    Class accepts dict, list, set, str, int, float, bool, None, dataclass instances
    """
    def __init__(self,
                 payload: Any,
                 status: HttpStatusCode = None,
                 headers: Optional[Dict[str, str]] = None,
                 charset: Charset = Charset.UTF8
                 ):
        try:
            payload = self.convert_payload(payload)
            payload = custom_dumps(payload)
        except TypeError as er:
            error_msg = f"Failed to serialise data: '{payload}' to json, problem: {er}"
            logger.exception(error_msg)
            payload = custom_dumps(error_msg)
            status = HttpStatusCode.InternalServerError

        super().__init__(payload, status, None, headers, charset)

    def convert_payload(self, payload):
        if payload_type := self.get_class_payload_type():
            if isinstance(payload, payload_type):
                return payload

            if dataclasses.is_dataclass(payload_type):
                # noinspection PyTypeChecker
                payload = dict_to_dataclass(payload, payload_type)
            else:
                # noinspection PyTypeChecker
                payload = convert_to_type(payload_type, '', payload)

        return payload

    @classmethod
    def get_class_content_type(cls) -> ContentType:
        return ContentType.Json

    @classmethod
    def get_class_payload_type(cls) -> type:
        if hasattr(cls, _FIXED_PAYLOAD_TYPE__NAME):
            return getattr(cls, _FIXED_PAYLOAD_TYPE__NAME)

    def __class_getitem__(cls, modifier: Union[HttpStatusCode, ContentType, type, None]) -> type:
        if isinstance(modifier, type):
            if cls.get_class_payload_type():
                raise TypeError("Payload type already fixed")
            return create_type(_prepare_class_name([cls], modifier), [cls], {_FIXED_PAYLOAD_TYPE__NAME: modifier})
        return super().__class_getitem__(modifier)


class WrappedAnswer(JsonableAnswer):
    def __init__(self, payload: Any = Nothing, status: HttpStatusCode = HttpStatusCode.OK,
                 headers: Optional[Dict[str, str]] = None, charset: Charset = Charset.UTF8) -> None:
        payload = {"done": True, "result": payload}
        super().__init__(payload, status, headers, charset)

    @classmethod
    def _wrap_type(cls, type_: type) -> type:
        attribute_key_to_type = {"done": bool, "result": type_} if type_ not  in (None, type[None]) else {"done": bool}
        return create_dataclass(_prepare_class_name([type_],  "Wrapped"), [cls], attribute_key_to_type)

    def __class_getitem__(cls, modifier: Union[HttpStatusCode, ContentType, type]) -> type:
        type_origin = getattr(modifier, '__origin__', None)
        if type_origin or isinstance(modifier, type):
            if cls.get_class_payload_type():
                raise TypeError("Payload type already fixed")
            wrapped_type = cls._wrap_type(modifier)
            return create_type(_prepare_class_name([cls], modifier), [cls], {_FIXED_PAYLOAD_TYPE__NAME: wrapped_type})
        return super().__class_getitem__(modifier)

    @staticmethod
    def _wrap_payload(payload: Any) -> Dict[str, Any]:
        return {"done": True, "result": payload}


@dataclasses.dataclass
class ErrorDescription:
    error: str
    error_type: Optional[str] = None
    error_code: Optional[ErrorCode] = None
    done: bool = False


class ErrorAnswer(JsonableAnswer):
    def __init__(self,
                 error: Any = Nothing,
                 status: HttpStatusCode = HttpStatusCode.InternalServerError,
                 error_type: Optional[str] = None,
                 error_code: Optional[ErrorCode] = None,
                 headers: Optional[Dict[str, str]] = None,
                 charset: Charset = Charset.UTF8) -> None:
        error_code = error_code or self.get_class_error_code()
        super().__init__(payload=ErrorDescription(error, error_type, error_code),
                         status=status,
                         headers=headers,
                         charset=charset,
                         )

    def __class_getitem__(cls, modifier: Union[HttpStatusCode, ContentType, ErrorCode, type, None]) -> type:
        if isinstance(modifier, ErrorCode):
            if cls.get_class_error_code():
                raise TypeError("ErrorCode already fixed")
            return create_type(_prepare_class_name([cls], modifier), [cls], {_FIXED_ERROR_CODE__NAME: modifier})
        return super().__class_getitem__(modifier)

    @classmethod
    def get_class_payload_type(cls) -> type:
        return ErrorDescription

    @classmethod
    def get_class_error_code(cls) -> Optional[ErrorCode]:
        if hasattr(cls, _FIXED_ERROR_CODE__NAME):
            return getattr(cls, _FIXED_ERROR_CODE__NAME)


class ExceptionAnswer(ErrorAnswer):
    def __init__(self,
                 exception: Exception,
                 status: Optional[HttpStatusCode] = None,
                 error_code: Optional[ErrorCode] = None,
                 headers: Optional[Dict[str, str]] = None,
                 charset: Charset = Charset.UTF8) -> None:
        super().__init__(error=str(exception),
                         status=status,
                         error_type=type(exception).__name__,
                         error_code=error_code,
                         headers=headers,
                         charset=charset,
                         )


class RedirectAnswer(Answer):
    def __init__(self, location: str, headers: Optional[Dict[str, str]] = None):
        headers = {} if headers is None else headers
        headers['Location'] = str(URL(location))
        super().__init__(None, HttpStatusCode.Found, headers=headers)

    @property
    def location(self):
        return self.headers['Location']


class FileAnswer(Answer):
    def __init__(self,
                 payload: Any,
                 file_name: str = None,
                 status: HttpStatusCode = HttpStatusCode.OK,
                 headers: dict[str, str] | None = None,
                 content_type: ContentType | str | None = None,
                 charset: str = None
                 ) -> None:
        if headers is None:
            headers = {}

        if file_name:
            file_name = urllib.parse.quote_plus(file_name)
            headers["Content-Disposition"] = f"attachment; filename=\"{file_name}\""

        if charset is None and content_type.startswith("text/"):
            charset = Charset.UTF8

        super().__init__(payload, status, content_type, headers, charset=charset)

==> http_tools_abm-5.4.65252.tar.gz: http_tools_abm-5.4.65252/http_tools/answer.py <==
import dataclasses
import logging
import urllib.parse
from typing import Any, Optional, Dict, Union

from aiohttp import Payload
from aiohttp.web_response import Response
from dynamic_types.class_name import _prepare_class_name
from dynamic_types.create import create_type, create_dataclass
from init_helpers import custom_dumps
from init_helpers.dict_to_dataclass import dict_to_dataclass, convert_to_type
from yarl import URL

from .charset import Charset
from .http_status_codes import HttpStatusCode
from .mime_types import ContentType

logger = logging.getLogger(__name__)


class Nothing:
    pass


class ErrorCode(int):
    pass


_FIXED_CONTENT_TYPE__NAME = "_fixed_content_type"
_FIXED_STATUS_CODE__NAME = "_fixed_status_code"
_FIXED_PAYLOAD_TYPE__NAME = "_fixed_payload_type"
_FIXED_ERROR_CODE__NAME = "_fixed_error_code_type"


class Answer(Response):
    def __init__(self,
                 payload: Union[str, bytes, None],
                 status: HttpStatusCode = None,
                 content_type: Optional[Union[ContentType, str]] = None,
                 headers: Optional[Dict[str, str]] = None,
                 charset: Charset = None,
                 ):
        if status is None and hasattr(self, _FIXED_STATUS_CODE__NAME):
            status = getattr(self, _FIXED_STATUS_CODE__NAME)

        status_code = status or self.get_class_content_type()
        content_type = content_type or self.get_class_content_type()
        content_type = content_type.value if isinstance(content_type, ContentType) else content_type
        super().__init__(body=payload, status=status_code, content_type=content_type, headers=headers, charset=charset)

    @property
    def bytes_body(self) -> bytes:
        return self.body._value if isinstance(self.body, Payload) else self.body

    def __class_getitem__(cls, modifier: Union[HttpStatusCode, ContentType]) -> type:
        field_name_to_value = {}
        if isinstance(modifier, HttpStatusCode):
            if cls.get_class_status_code():
                raise TypeError("Status code already fixed")
            field_name_to_value[_FIXED_STATUS_CODE__NAME] = modifier
        elif isinstance(modifier, ContentType):
            if cls.get_class_content_type():
                raise TypeError("Content type already fixed")
            field_name_to_value[_FIXED_CONTENT_TYPE__NAME] = modifier
        else:
            raise TypeError(f"Unexpected modifier type: {type(modifier)}")

        new_type = create_type(_prepare_class_name([cls], modifier), [cls], field_name_to_value)
        # new_type = type(_prepare_class_name([cls], modifier), (cls, ), field_name_to_value)
        return new_type

    @classmethod
    def get_class_content_type(cls) -> Optional[ContentType]:
        if hasattr(cls, _FIXED_CONTENT_TYPE__NAME):
            return getattr(cls, _FIXED_CONTENT_TYPE__NAME)

    @classmethod
    def get_class_status_code(cls) -> Optional[HttpStatusCode]:
        if hasattr(cls, _FIXED_STATUS_CODE__NAME):
            return getattr(cls, _FIXED_STATUS_CODE__NAME)

    @classmethod
    def get_class_payload_type(cls) -> type:
        return bytes


class JsonableAnswer(Answer):
    """
    Class accepts dict, list, set, str, int, float, bool, None, dataclass instances
    """
    def __init__(self,
                 payload: Any,
                 status: HttpStatusCode = None,
                 headers: Optional[Dict[str, str]] = None,
                 charset: Charset = Charset.UTF8
                 ):
        try:
            payload = self.convert_payload(payload)
            payload = custom_dumps(payload)
        except TypeError as er:
            error_msg = f"Failed to serialise data: '{payload}' to json, problem: {er}"
            logger.exception(error_msg)
            payload = custom_dumps(error_msg)
            status = HttpStatusCode.InternalServerError

        super().__init__(payload, status, None, headers, charset)

    def convert_payload(self, payload):
        if payload_type := self.get_class_payload_type():
            if isinstance(payload, payload_type):
                return payload

            if dataclasses.is_dataclass(payload_type):
                # noinspection PyTypeChecker
                payload = dict_to_dataclass(payload, payload_type)
            else:
                # noinspection PyTypeChecker
                payload = convert_to_type(payload_type, '', payload)

        return payload

    @classmethod
    def get_class_content_type(cls) -> ContentType:
        return ContentType.Json

    @classmethod
    def get_class_payload_type(cls) -> type:
        if hasattr(cls, _FIXED_PAYLOAD_TYPE__NAME):
            return getattr(cls, _FIXED_PAYLOAD_TYPE__NAME)

    def __class_getitem__(cls, modifier: Union[HttpStatusCode, ContentType, type, None]) -> type:
        if isinstance(modifier, type):
            if cls.get_class_payload_type():
                raise TypeError("Payload type already fixed")
            return create_type(_prepare_class_name([cls], modifier), [cls], {_FIXED_PAYLOAD_TYPE__NAME: modifier})
        return super().__class_getitem__(modifier)


class WrappedAnswer(JsonableAnswer):
    def __init__(self, payload: Any = Nothing, status: HttpStatusCode = HttpStatusCode.OK,
                 headers: Optional[Dict[str, str]] = None, charset: Charset = Charset.UTF8) -> None:
        payload = {"done": True, "result": payload}
        super().__init__(payload, status, headers, charset)

    @classmethod
    def _wrap_type(cls, type_: type) -> type:
        attribute_key_to_type = {"done": bool, "result": type_} if type_ not  in (None, type[None]) else {"done": bool}
        return create_dataclass(_prepare_class_name([type_],  "Wrapped"), [cls], attribute_key_to_type)

    def __class_getitem__(cls, modifier: Union[HttpStatusCode, ContentType, type]) -> type:
        type_origin = getattr(modifier, '__origin__', None)
        if type_origin or isinstance(modifier, type):
            if cls.get_class_payload_type():
                raise TypeError("Payload type already fixed")
            wrapped_type = cls._wrap_type(modifier)
            return create_type(_prepare_class_name([cls], modifier), [cls], {_FIXED_PAYLOAD_TYPE__NAME: wrapped_type})
        return super().__class_getitem__(modifier)

    @staticmethod
    def _wrap_payload(payload: Any) -> Dict[str, Any]:
        return {"done": True, "result": payload}


@dataclasses.dataclass
class ErrorDescription:
    error: str
    error_type: Optional[str] = None
    error_code: Optional[ErrorCode] = None
    done: bool = False


class ErrorAnswer(JsonableAnswer):
    def __init__(self,
                 error: Any = Nothing,
                 status: HttpStatusCode = HttpStatusCode.InternalServerError,
                 error_type: Optional[str] = None,
                 error_code: Optional[ErrorCode] = None,
                 headers: Optional[Dict[str, str]] = None,
                 charset: Charset = Charset.UTF8) -> None:
        error_code = error_code or self.get_class_error_code()
        super().__init__(payload=ErrorDescription(error, error_type, error_code),
                         status=status,
                         headers=headers,
                         charset=charset,
                         )

    def __class_getitem__(cls, modifier: Union[HttpStatusCode, ContentType, ErrorCode, type, None]) -> type:
        if isinstance(modifier, ErrorCode):
            if cls.get_class_error_code():
                raise TypeError("ErrorCode already fixed")
            return create_type(_prepare_class_name([cls], modifier), [cls], {_FIXED_ERROR_CODE__NAME: modifier})
        return super().__class_getitem__(modifier)

    @classmethod
    def get_class_payload_type(cls) -> type:
        return ErrorDescription

    @classmethod
    def get_class_error_code(cls) -> Optional[ErrorCode]:
        if hasattr(cls, _FIXED_ERROR_CODE__NAME):
            return getattr(cls, _FIXED_ERROR_CODE__NAME)


class ExceptionAnswer(ErrorAnswer):
    def __init__(self,
                 exception: Exception,
                 status: Optional[HttpStatusCode] = None,
                 error_code: Optional[ErrorCode] = None,
                 headers: Optional[Dict[str, str]] = None,
                 charset: Charset = Charset.UTF8) -> None:
        super().__init__(error=str(exception),
                         status=status,
                         error_type=type(exception).__name__,
                         error_code=error_code,
                         headers=headers,
                         charset=charset,
                         )


class RedirectAnswer(Answer):
    def __init__(self, location: str, headers: Optional[Dict[str, str]] = None):
        headers = {} if headers is None else headers
        headers['Location'] = str(URL(location))
        super().__init__(None, HttpStatusCode.Found, headers=headers)

    @property
    def location(self):
        return self.headers['Location']


class FileAnswer(Answer):
    def __init__(self,
                 payload: Any,
                 file_name: str = None,
                 status: HttpStatusCode = HttpStatusCode.OK,
                 headers: dict[str, str] | None = None,
                 content_type: ContentType | str | None = None,
                 charset: str = None
                 ) -> None:
        if headers is None:
            headers = {}

        if file_name:
            file_name = urllib.parse.quote_plus(file_name)
            headers["Content-Disposition"] = f"attachment; filename=\"{file_name}\""

        content_type = content_type or self.get_class_content_type()
        if charset is None and content_type is not None and content_type.startswith("text/"):
            charset = Charset.UTF8

        super().__init__(payload, status, content_type, headers, charset=charset)
